sdk_path = "/home/vagrant/android-sdk"
ndk_paths = {
    'r10e': "/home/vagrant/android-ndk/r10e",
    'r11c': "/home/vagrant/android-ndk/r11c",
    'r12b': "/home/vagrant/android-ndk/r12b",
    'r13b': "/home/vagrant/android-ndk/r13b",
    'r14b': "/home/vagrant/android-ndk/r14b",
    'r15c': "/home/vagrant/android-ndk/r15c",
    'r16b': "/home/vagrant/android-ndk/r16b",
}
java_paths = {
    '8': "/usr/lib/jvm/java-8-openjdk-amd64",
}
